"""DatasetSpec registry (single source of truth for validation + benchmark datasets)."""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing import Any, Dict, Optional, Tuple

@dataclass(frozen=True)
class DatasetSpec:
    dataset_id: str
    display_name: str
    pipeline: str
    tier: str
    loader_kind: str
    params: Dict[str, Any] = field(default_factory=dict)
    # Registry/benchmark metadata (filled in finalize step)
    source_kind: str = "validation_catalog"  # synthetic | validation_catalog | dist_benchmark
    domain: str = "genomics"  # genomics | non_genomic | synthetic
    platform: str = "cDNA"  # Affy HG-U95 | Affy HG-U133A | cDNA | mass-spec | synthetic
    sota_holdout_bal_acc: Optional[Tuple[float, float]] = None
    sota_inflated_bal_acc: Optional[Tuple[float, float]] = None
    provenance: str = ""
    extended_only: bool = False  # True = only included in the expanded/extended catalog


def _build_dataset_registry() -> Dict[str, DatasetSpec]:
    fs_specs = {
        "leukemia_golub": DatasetSpec(
            dataset_id="leukemia_golub",
            display_name="Leukemia (Golub)",
            pipeline="fs",
            tier="easy",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "leukemia", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 72,
                    "n_features": 7129,
                    "n_classes": 2,
                    # OpenML `leukemia` v1: 47/25 class split.
                    "weights": [47 / 72.0, 25 / 72.0],
                    "difficulty": "easy",
                },
                "n_final_features": 50,
            },
        ),
        "dlbcl_shipp": DatasetSpec(
            dataset_id="dlbcl_shipp",
            display_name="DLBCL (Shipp)",
            pipeline="fs",
            tier="easy",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "DLBCL", "version": 1},
                    {"name": "dlbcl", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 77,
                    # OpenML `DLBCL` v1 uses 5,469 genes in the published-preprocessed variant.
                    "n_features": 5469,
                    "n_classes": 2,
                    "weights": [58 / 77.0, 19 / 77.0],
                    "difficulty": "easy",
                },
                "n_final_features": 50,
            },
        ),
        "ovarian_petricoin": DatasetSpec(
            dataset_id="ovarian_petricoin",
            display_name="Ovarian Cancer (Petricoin)",
            pipeline="fs",
            tier="easy",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "ovarian", "version": 1},
                    {"name": "Ovarian", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 253,
                    "n_features": 15154,
                    "n_classes": 2,
                    "weights": [162 / 253.0, 91 / 253.0],
                    "difficulty": "easy",
                },
                "n_final_features": 60,
            },
        ),
        "srbct_khan": DatasetSpec(
            dataset_id="srbct_khan",
            display_name="SRBCT (Khan)",
            pipeline="fs",
            tier="easy",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "SRBCT", "version": 1},
                    {"name": "srbct", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 83,
                    "n_features": 2308,
                    "n_classes": 4,
                    # OpenML `SRBCT` v1: [29, 11, 18, 25] class split.
                    "weights": [29 / 83.0, 11 / 83.0, 18 / 83.0, 25 / 83.0],
                    "difficulty": "easy",
                },
                "n_final_features": 50,
            },
        ),
        "prostate_singh": DatasetSpec(
            dataset_id="prostate_singh",
            display_name="Prostate Cancer (Singh)",
            pipeline="fs",
            tier="easy",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "prostate", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 102,
                    "n_features": 12600,
                    "n_classes": 2,
                    "weights": [52 / 102.0, 50 / 102.0],
                    "difficulty": "easy",
                },
                "n_final_features": 50,
            },
        ),
        "mll_microarray": DatasetSpec(
            dataset_id="mll_microarray",
            display_name="MLL (OpenML)",
            pipeline="fs",
            tier="easy",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "MLL", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 72,
                    "n_features": 12582,
                    "n_classes": 3,
                    # OpenML `MLL` v1: 24/28/20 class split.
                    "weights": [24 / 72.0, 28 / 72.0, 20 / 72.0],
                    "difficulty": "easy",
                },
                "n_final_features": 60,
            },
        ),
        "gli_85": DatasetSpec(
            dataset_id="gli_85",
            display_name="Glioma (GLI_85)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "GLI", "version": 1},
                ],
                "synthetic_profile": {
                    # OpenML `GLI` v1: (85, 22283), binary.
                    "n_samples": 85,
                    "n_features": 22283,
                    "n_classes": 2,
                    # OpenML target split: 26 / 59.
                    "weights": [26 / 85.0, 59 / 85.0],
                    "difficulty": "medium",
                },
                "n_final_features": 60,
            },
        ),
        "smk_can_187": DatasetSpec(
            dataset_id="smk_can_187",
            display_name="SMK_CAN_187 (Smoking)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "SMK", "version": 1},
                ],
                "synthetic_profile": {
                    # OpenML `SMK` v1: (187, 19993), binary.
                    "n_samples": 187,
                    "n_features": 19993,
                    "n_classes": 2,
                    # OpenML target split: 90 / 97.
                    "weights": [90 / 187.0, 97 / 187.0],
                    "difficulty": "medium",
                },
                "n_final_features": 60,
            },
        ),
        # --- Scikit-feature/ASU expansion candidates (2026-02): multiclass HDLSS ---
        "cll_sub_111": DatasetSpec(
            dataset_id="cll_sub_111",
            display_name="CLL_SUB_111 (Scikit-feature)",
            pipeline="fs",
            tier="medium",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/CLL-SUB-111.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/CLL-SUB-111.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 111,
                    "n_features": 11340,
                    "n_classes": 3,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
        ),
        "tox_171": DatasetSpec(
            dataset_id="tox_171",
            display_name="TOX_171 (Scikit-feature)",
            pipeline="fs",
            tier="medium",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/TOX-171.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/TOX-171.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 171,
                    "n_features": 5748,
                    "n_classes": 4,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
        ),
        "gla_bra_180": DatasetSpec(
            dataset_id="gla_bra_180",
            display_name="GLA-BRA-180 (Scikit-feature)",
            pipeline="fs",
            tier="hard",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/GLA-BRA-180.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/GLA-BRA-180.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 180,
                    "n_features": 49151,
                    "n_classes": 4,
                    "difficulty": "hard",
                },
                "n_final_features": 120,
            },
        ),
        "carcinom_11class": DatasetSpec(
            dataset_id="carcinom_11class",
            display_name="Carcinom 11-class (Scikit-feature)",
            pipeline="fs",
            tier="hard",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/Carcinom.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 174,
                    "n_features": 9182,
                    "n_classes": 11,
                    "difficulty": "hard",
                },
                "n_final_features": 90,
            },
        ),
        "glioma_50_4class": DatasetSpec(
            dataset_id="glioma_50_4class",
            display_name="Glioma (50 x 4434, 4-class)",
            pipeline="fs",
            tier="medium",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/GLIOMA.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 50,
                    "n_features": 4434,
                    "n_classes": 4,
                    "difficulty": "medium",
                },
                "n_final_features": 70,
            },
        ),
        # Nutt et al. 2003, Cancer Research 63(7):1602-1607.
        # 50 samples, 12625 genes (Affymetrix U95Av2), 4 classes: CG/CO/NG/NO.
        # Source: biolab.si (University of Ljubljana / Orange Data Mining).
        "brain_tumor_2_50_4class": DatasetSpec(
            dataset_id="brain_tumor_2_50_4class",
            display_name="Brain Tumor 2 — Nutt 2003 (50 x 12625, 4-class)",
            pipeline="fs",
            tier="medium",
            loader_kind="tab_url_or_synth",
            params={
                "tab_url_options": [
                    {
                        "url": "https://file.biolab.si/biolab/supp/bi-cancer/projections/_datasets/glioblastoma.tab",
                    },
                ],
                "openml_options": [
                    {"name": "Brain_Tumor2", "version": 1},
                    {"name": "Brain_Tumor_2", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 50,
                    "n_features": 12625,
                    "n_classes": 4,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
        ),
        # Armstrong et al. 2002, Nature Genetics 30(1):41-47 (MLL leukemia).
        # 72 samples, 12533 genes (Affymetrix U95Av2), 3 classes: ALL/AML/MLL.
        # Source: biolab.si (University of Ljubljana / Orange Data Mining).
        "leukemia_1_72_3class": DatasetSpec(
            dataset_id="leukemia_1_72_3class",
            display_name="MLL Leukemia — Armstrong 2002 (72 x 12533, 3-class)",
            pipeline="fs",
            tier="medium",
            loader_kind="tab_url_or_synth",
            params={
                "tab_url_options": [
                    {
                        "url": "https://file.biolab.si/biolab/supp/bi-cancer/projections/_datasets/MLL.tab",
                    },
                ],
                "openml_options": [
                    {"name": "Leukemia1", "version": 1},
                    {"name": "leukemia1", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 72,
                    "n_features": 12533,
                    "n_classes": 3,
                    "difficulty": "medium",
                },
                "n_final_features": 70,
            },
        ),
        # NOTE: A legacy duplicate leukemia variant was removed from the catalog.
        # (same Armstrong 2002 MLL data with a different feature-filtering threshold
        # from the GEMS benchmark, which is now dead: gems-system.org).
        "nci9_60_9class": DatasetSpec(
            dataset_id="nci9_60_9class",
            display_name="NCI9 (60 x 9712, 9-class)",
            pipeline="fs",
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "NCI9", "version": 1},
                    {"name": "9_Tumors", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 60,
                    "n_features": 9712,
                    "n_classes": 9,
                    "difficulty": "hard",
                },
                "n_final_features": 80,
            },
        ),
        "nci_61_8class": DatasetSpec(
            dataset_id="nci_61_8class",
            display_name="NCI (61 x 5244, 8-class proxy)",
            pipeline="fs",
            tier="hard",
            loader_kind="nci60_proxy_or_synth",
            params={
                # OpenML no longer exposes the legacy NCI-61 dataset. We build a
                # deterministic real-data proxy from ISLR NCI60 by:
                # - mapping K562* to LEUKEMIA and MCF7* to BREAST,
                # - dropping UNKNOWN and PROSTATE,
                # - retaining top-variance genes to 5,244 features.
                "proxy_target_features": 5244,
                "synthetic_profile": {
                    "n_samples": 61,
                    "n_features": 5244,
                    "n_classes": 8,
                    "difficulty": "hard",
                },
                "n_final_features": 80,
            },
        ),
        # --- Non-genomics HDLSS additions used in recent FS papers ---
        "orlraws10p": DatasetSpec(
            dataset_id="orlraws10p",
            display_name="ORLraws10P (Face)",
            pipeline="fs",
            tier="easy",
            loader_kind="face_proxy_or_synth",
            params={
                "domain": "face",
                "face_resize_shape": (112, 92),
                "synthetic_profile": {
                    "n_samples": 100,
                    "n_features": 10304,
                    "n_classes": 10,
                    "difficulty": "easy",
                },
                "n_final_features": 90,
            },
        ),
        "warp_pie10p": DatasetSpec(
            dataset_id="warp_pie10p",
            display_name="warpPIE10P (Face)",
            pipeline="fs",
            tier="easy",
            loader_kind="face_proxy_or_synth",
            params={
                "domain": "face",
                "face_resize_shape": (55, 44),
                "synthetic_profile": {
                    "n_samples": 210,
                    "n_features": 2420,
                    "n_classes": 10,
                    "difficulty": "easy",
                },
                "n_final_features": 80,
            },
        ),
        "pixraw10p": DatasetSpec(
            dataset_id="pixraw10p",
            display_name="pixraw10P (Face)",
            pipeline="fs",
            tier="easy",
            loader_kind="face_proxy_or_synth",
            params={
                "domain": "face",
                "face_resize_shape": (100, 100),
                "synthetic_profile": {
                    "n_samples": 100,
                    "n_features": 10000,
                    "n_classes": 10,
                    "difficulty": "easy",
                },
                "n_final_features": 90,
            },
        ),
        "hf_breast_ge_mubashir1837": DatasetSpec(
            dataset_id="hf_breast_ge_mubashir1837",
            display_name="Breast Cancer Gene Expression (HF: mubashir1837)",
            pipeline="fs",
            # Routed to hard: tiny-n (51) HDLSS RNA-seq benchmark with no protocol-stable
            # published SOTA under strict holdout (placeholder strict range ~0.50–0.70).
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "local_path_env": "HF_BREAST_GE_PATH",
                "default_local_path": "train_data/huggingface/mubashir1837_breast_cancer_ge/cleaned_expression.csv",
                "synthetic_profile": {
                    # HuggingFace dataset: 51 samples, ~28k genes, binary response label.
                    "n_samples": 51,
                    "n_features": 28278,
                    "n_classes": 2,
                    "weights": [26 / 51.0, 25 / 51.0],
                    "difficulty": "hard",
                },
                "n_final_features": 80,
                "source_policy": "real_only",
            },
        ),
        "lymphoma_3": DatasetSpec(
            dataset_id="lymphoma_3",
            display_name="Lymphoma-3 (OpenML)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "Lymphoma-3", "version": 1},
                ],
                "synthetic_profile": {
                    # OpenML `Lymphoma-3` v1: (66, 4026), 3 classes.
                    "n_samples": 66,
                    "n_features": 4026,
                    "n_classes": 3,
                    # OpenML target split: 11 / 46 / 9.
                    "weights": [11 / 66.0, 46 / 66.0, 9 / 66.0],
                    "difficulty": "medium",
                },
                "n_final_features": 60,
            },
        ),
        "lymphoma_9": DatasetSpec(
            dataset_id="lymphoma_9",
            display_name="Lymphoma-9 (OpenML)",
            pipeline="fs",
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "Lymphoma-9", "version": 1},
                ],
                "synthetic_profile": {
                    # OpenML `Lymphoma-9` v1: (96, 4026), 9 classes.
                    "n_samples": 96,
                    "n_features": 4026,
                    "n_classes": 9,
                    # OpenML target split: 10/11/46/9/2/2/6/4/6 (ABB/CLL/DLBCL/FL/GCB/NIL/RAT/RBB/TCL).
                    "weights": [
                        10 / 96.0,
                        11 / 96.0,
                        46 / 96.0,
                        9 / 96.0,
                        2 / 96.0,
                        2 / 96.0,
                        6 / 96.0,
                        4 / 96.0,
                        6 / 96.0,
                    ],
                    "difficulty": "hard",
                },
                "n_final_features": 70,
            },
        ),
        "lymphoma_11": DatasetSpec(
            dataset_id="lymphoma_11",
            display_name="Lymphoma-11 (OpenML)",
            pipeline="fs",
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "Lymphoma-11", "version": 1},
                ],
                "synthetic_profile": {
                    # OpenML `Lymphoma-11` v1: (96, 4026), 11 classes.
                    "n_samples": 96,
                    "n_features": 4026,
                    "n_classes": 11,
                    # OpenML target split: 10/23/11/1/9/2/22/2/6/4/6
                    # (ABB/ACL/CLL/DLBCL/FL/GCB/GCL/NIL/RAT/RBB/TCL).
                    "weights": [
                        10 / 96.0,
                        23 / 96.0,
                        11 / 96.0,
                        1 / 96.0,
                        9 / 96.0,
                        2 / 96.0,
                        22 / 96.0,
                        2 / 96.0,
                        6 / 96.0,
                        4 / 96.0,
                        6 / 96.0,
                    ],
                    "difficulty": "hard",
                },
                "n_final_features": 70,
            },
        ),
        "tumor9_openml": DatasetSpec(
            dataset_id="tumor9_openml",
            display_name="9-Tumors (OpenML)",
            pipeline="fs",
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "9_Tumors", "version": 1},
                ],
                "synthetic_profile": {
                    # OpenML `9_Tumors` v1: (60, 5726), 9 classes.
                    "n_samples": 60,
                    "n_features": 5726,
                    "n_classes": 9,
                    "difficulty": "hard",
                },
                "n_final_features": 60,
            },
        ),
        "arcene_nips03": DatasetSpec(
            dataset_id="arcene_nips03",
            display_name="ARCENE (NIPS 2003 FS Challenge)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "arcene", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 200,
                    "n_features": 10000,
                    "n_classes": 2,
                    "weights": [0.5, 0.5],
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
        ),
        "madelon_nips03": DatasetSpec(
            dataset_id="madelon_nips03",
            display_name="MADELON (NIPS 2003 FS Challenge)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "madelon", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 2600,
                    "n_features": 500,
                    "n_classes": 2,
                    "weights": [0.5, 0.5],
                    "difficulty": "medium",
                },
                "n_final_features": 50,
            },
        ),
        "gisette_nips03": DatasetSpec(
            dataset_id="gisette_nips03",
            display_name="GISETTE (NIPS 2003 FS Challenge)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    # OpenML `gisette` v1 is inactive/retired; prefer active v2 mirror.
                    {"name": "gisette", "version": 2},
                    {"data_id": 41026},
                ],
                "synthetic_profile": {
                    "n_samples": 7000,
                    "n_features": 5001,
                    "n_classes": 2,
                    "weights": [0.5, 0.5],
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
        ),
        "dexter_nips03": DatasetSpec(
            dataset_id="dexter_nips03",
            display_name="DEXTER (NIPS 2003 FS Challenge)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "dexter", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 600,
                    "n_features": 20001,
                    "n_classes": 2,
                    "weights": [0.5, 0.5],
                    "difficulty": "medium",
                },
                "openml_feature_cap": 20001,
                "n_final_features": 120,
                "source_policy": "real_only",
            },
        ),
        "dorothea_nips03": DatasetSpec(
            dataset_id="dorothea_nips03",
            display_name="DOROTHEA (NIPS 2003 FS Challenge)",
            pipeline="fs",
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"data_id": 4137},
                    {"name": "dorothea", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 1150,
                    "n_features": 100001,
                    "n_classes": 2,
                    # OpenML `Dorothea` v1: 1038/112 class split.
                    "weights": [1038 / 1150.0, 112 / 1150.0],
                    "difficulty": "hard",
                },
                # Dorothea is extremely sparse and very wide; cap features before densifying.
                "openml_feature_cap": 25000,
                "n_final_features": 200,
            },
        ),
        "colon_alon": DatasetSpec(
            dataset_id="colon_alon",
            display_name="Colon Cancer (Alon)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "colon", "version": 1},
                    {"name": "Colon", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 62,
                    "n_features": 2000,
                    "n_classes": 2,
                    "weights": [40 / 62.0, 22 / 62.0],
                    "difficulty": "medium",
                },
                "n_final_features": 40,
            },
        ),
        "cns_pomeroy": DatasetSpec(
            dataset_id="cns_pomeroy",
            display_name="CNS / Brain Tumors (Pomeroy)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "CNS", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 60,
                    "n_features": 7129,
                    "n_classes": 2,
                    # OpenML `CNS` v1: 39/21 class split.
                    "weights": [39 / 60.0, 21 / 60.0],
                    "difficulty": "medium",
                },
                "n_final_features": 45,
            },
        ),
        "lung_gordon": DatasetSpec(
            dataset_id="lung_gordon",
            # NOTE: `lung_gordon` is a legacy id retained for compatibility.
            # The automatically-loadable OpenML dataset is a 5-class lung benchmark.
            display_name="Lung Cancer (OpenML Lung, 5-class)",
            pipeline="fs",
            # Routed to hard: 5-class with a 6-sample minority class makes strict 80/20
            # holdout gate-unreliable (single-error ΔBA ≈ 0.20 when n_test(min)=1).
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "lung", "version": 1},
                ],
                "synthetic_profile": {
                    # OpenML `Lung` v1: (203, 12600), 5 classes with heavy imbalance.
                    "n_samples": 203,
                    "n_features": 12600,
                    "n_classes": 5,
                    "weights": [139 / 203.0, 17 / 203.0, 6 / 203.0, 21 / 203.0, 20 / 203.0],
                    "difficulty": "hard",
                },
                "n_final_features": 60,
            },
        ),
        "breast_vantveer": DatasetSpec(
            dataset_id="breast_vantveer",
            display_name="Breast Cancer Prognosis (van 't Veer)",
            pipeline="fs",
            tier="medium",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    # Microarray-tagged OpenML dataset; matches (97, 24481) profile.
                    {"name": "Breast", "version": 3},
                ],
                "synthetic_profile": {
                    "n_samples": 97,
                    "n_features": 24481,
                    "n_classes": 2,
                    "weights": [46 / 97.0, 51 / 97.0],
                    "difficulty": "medium",
                },
                "n_final_features": 60,
            },
        ),
        "nci60_ross": DatasetSpec(
            dataset_id="nci60_ross",
            display_name="NCI60 (Ross)",
            pipeline="fs",
            tier="hard",
            loader_kind="nci60_url_or_synth",
            params={
                "synthetic_profile": {
                    "n_samples": 60,
                    "n_features": 6830,
                    "n_classes": 9,
                    "weights": [0.18, 0.15, 0.12, 0.12, 0.11, 0.1, 0.09, 0.08, 0.05],
                    "difficulty": "hard",
                },
                "n_final_features": 50,
            },
        ),
        "gcm_ramaswamy": DatasetSpec(
            dataset_id="gcm_ramaswamy",
            display_name="GCM (Ramaswamy)",
            pipeline="fs",
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "GCM", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 198,
                    "n_features": 16063,
                    "n_classes": 14,
                    "weights": [
                        0.11,
                        0.1,
                        0.09,
                        0.08,
                        0.08,
                        0.07,
                        0.07,
                        0.07,
                        0.06,
                        0.06,
                        0.06,
                        0.05,
                        0.05,
                        0.05,
                    ],
                    "difficulty": "hard",
                },
                "n_final_features": 70,
            },
        ),
        "tumor11_su": DatasetSpec(
            dataset_id="tumor11_su",
            display_name="11-Tumor (Su)",
            pipeline="fs",
            tier="hard",
            loader_kind="openml_or_synth",
            params={
                "openml_options": [
                    {"name": "11_Tumors", "version": 1},
                ],
                "synthetic_profile": {
                    "n_samples": 174,
                    "n_features": 12533,
                    "n_classes": 11,
                    "weights": [
                        0.13,
                        0.12,
                        0.1,
                        0.1,
                        0.09,
                        0.09,
                        0.08,
                        0.08,
                        0.07,
                        0.07,
                        0.07,
                    ],
                    "difficulty": "hard",
                },
                "n_final_features": 65,
            },
        ),
        "nci60_strict_holdout": DatasetSpec(
            dataset_id="nci60_strict_holdout",
            display_name="NCI60 Strict Holdout",
            pipeline="fs",
            tier="very_hard",
            loader_kind="nci60_url_or_synth",
            params={
                "synthetic_profile": {
                    "n_samples": 60,
                    "n_features": 6830,
                    "n_classes": 9,
                    "weights": [0.2, 0.15, 0.13, 0.12, 0.1, 0.09, 0.08, 0.07, 0.06],
                    "difficulty": "very_hard",
                },
                "fs_fraction": 0.4,
                "n_final_features": 24,
            },
        ),
        "cumida_leukemia_subtypes": DatasetSpec(
            dataset_id="cumida_leukemia_subtypes",
            display_name="CuMiDa Leukemia Subtypes",
            pipeline="fs",
            tier="very_hard",
            loader_kind="manual_or_synth",
            params={
                "local_path_env": "CUMIDA_LEUKEMIA_PATH",
                "default_local_path": "train_data/cumida/Leukemia_GSE28497.arff",
                "synthetic_profile": {
                    # CuMiDa Leukemia GSE28497 export shipped in-repo.
                    "n_samples": 281,
                    "n_features": 22283,
                    "n_classes": 7,
                    "weights": [0.24, 0.2, 0.16, 0.14, 0.1, 0.09, 0.07],
                    "difficulty": "very_hard",
                },
                "n_final_features": 70,
            },
        ),
        "cumida_brain_gse50161": DatasetSpec(
            dataset_id="cumida_brain_gse50161",
            display_name="CuMiDa Brain Subtypes (GSE50161)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "local_path_env": "CUMIDA_BRAIN_PATH",
                "default_local_path": "train_data/cumida/Brain_GSE50161.arff",
                # Real-data profile reference:
                # 108 valid samples, 4 classes (ependymoma=46, glioblastoma=34,
                # medulloblastoma=22, normal=6).
                "synthetic_profile": {
                    "n_samples": 108,
                    "n_features": 54675,
                    "n_classes": 4,
                    "weights": [46 / 108.0, 34 / 108.0, 22 / 108.0, 6 / 108.0],
                    "difficulty": "hard",
                },
                "n_final_features": 120,
            },
        ),
        "cumida_breast_gse45827": DatasetSpec(
            dataset_id="cumida_breast_gse45827",
            display_name="CuMiDa Breast Subtypes (GSE45827)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "local_path_env": "CUMIDA_BREAST_PATH",
                "default_local_path": "train_data/cumida/Breast_GSE45827.arff",
                "synthetic_profile": {
                    # CuMiDa Breast GSE45827: 151 samples, 54,675 probes, 6 classes
                    # (basal, HER, cell_line, normal, luminal_A, luminal_B).
                    "n_samples": 151,
                    "n_features": 54675,
                    "n_classes": 6,
                    # Class counts from CuMiDa `Breast_GSE45827_classes.txt`:
                    # basal=41, HER=30, cell_line=14, normal=7, luminal_A=29, luminal_B=30.
                    "weights": [
                        41 / 151.0,
                        30 / 151.0,
                        14 / 151.0,
                        7 / 151.0,
                        29 / 151.0,
                        30 / 151.0,
                    ],
                    "difficulty": "hard",
                },
                "n_final_features": 140,
            },
        ),
        # ---------------------------------------------------------------
        # Extended-only CuMiDa datasets (not in original validation bundle)
        # Each uses a pre-compiled NPZ file in train_data/extended/
        # ---------------------------------------------------------------
        "cumida_prostate_gse6919": DatasetSpec(
            dataset_id="cumida_prostate_gse6919",
            display_name="CuMiDa Prostate (GSE6919)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_prostate_gse6919.npz",
                "synthetic_profile": {
                    "n_samples": 171,
                    "n_features": 12625,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "cumida_ovarian_gse26712": DatasetSpec(
            dataset_id="cumida_ovarian_gse26712",
            display_name="CuMiDa Ovarian (GSE26712)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_ovarian_gse26712.npz",
                "synthetic_profile": {
                    "n_samples": 195,
                    "n_features": 22283,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "cumida_lung_gse19804": DatasetSpec(
            dataset_id="cumida_lung_gse19804",
            display_name="CuMiDa Lung (GSE19804)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_lung_gse19804.npz",
                "synthetic_profile": {
                    "n_samples": 120,
                    "n_features": 54675,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "cumida_colorectal_gse44861": DatasetSpec(
            dataset_id="cumida_colorectal_gse44861",
            display_name="CuMiDa Colorectal (GSE44861)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_colorectal_gse44861.npz",
                "synthetic_profile": {
                    "n_samples": 111,
                    "n_features": 22277,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "cumida_gastric_gse54129": DatasetSpec(
            dataset_id="cumida_gastric_gse54129",
            display_name="CuMiDa Gastric (GSE54129)",
            pipeline="fs",
            tier="easy",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_gastric_gse54129.npz",
                "synthetic_profile": {
                    "n_samples": 132,
                    "n_features": 54675,
                    "n_classes": 2,
                    "difficulty": "easy",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "cumida_pancreatic_gse16515": DatasetSpec(
            dataset_id="cumida_pancreatic_gse16515",
            display_name="CuMiDa Pancreatic (GSE16515)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_pancreatic_gse16515.npz",
                "synthetic_profile": {
                    "n_samples": 52,
                    "n_features": 54613,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "cumida_renal_gse53757": DatasetSpec(
            dataset_id="cumida_renal_gse53757",
            display_name="CuMiDa Renal (GSE53757)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_renal_gse53757.npz",
                "synthetic_profile": {
                    "n_samples": 144,
                    "n_features": 54675,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "cumida_headneck_gse12452": DatasetSpec(
            dataset_id="cumida_headneck_gse12452",
            display_name="CuMiDa Head/Neck (GSE12452)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/cumida_headneck_gse12452.npz",
                "synthetic_profile": {
                    "n_samples": 41,
                    "n_features": 54675,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        # ---------------------------------------------------------------
        # Extended-only UCSC Xena / TCGA datasets (gene expression, tabular)
        # Pre-compiled NPZ files in train_data/extended/ from UCSC Xena TCGA hub
        # ---------------------------------------------------------------
        "xena_tcga_brca": DatasetSpec(
            dataset_id="xena_tcga_brca",
            display_name="TCGA-BRCA Breast Cancer (Xena, PAM50 5-subtype)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_brca.npz",
                "synthetic_profile": {
                    "n_samples": 956,
                    "n_features": 20530,
                    "n_classes": 5,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_luad": DatasetSpec(
            dataset_id="xena_tcga_luad",
            display_name="TCGA-LUAD Lung Adenocarcinoma (Xena, 3-subtype)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_luad.npz",
                "synthetic_profile": {
                    "n_samples": 275,
                    "n_features": 20530,
                    "n_classes": 3,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_ucec": DatasetSpec(
            dataset_id="xena_tcga_ucec",
            display_name="TCGA-UCEC Uterine (Xena, histological 3-type)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_ucec.npz",
                "synthetic_profile": {
                    "n_samples": 190,
                    "n_features": 20530,
                    "n_classes": 3,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_lgg": DatasetSpec(
            dataset_id="xena_tcga_lgg",
            display_name="TCGA-LGG Lower Grade Glioma (Xena, 3-subtype)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_lgg.npz",
                "synthetic_profile": {
                    "n_samples": 529,
                    "n_features": 20530,
                    "n_classes": 3,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_kirc": DatasetSpec(
            dataset_id="xena_tcga_kirc",
            display_name="TCGA-KIRC Kidney Clear Cell (Xena, stage-based)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_kirc.npz",
                "synthetic_profile": {
                    "n_samples": 606,
                    "n_features": 20530,
                    "n_classes": 4,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        # ---------------------------------------------------------------
        # Extended-only UCSC Xena / TCGA expansion (Phase 2)
        # ---------------------------------------------------------------
        "xena_tcga_hnsc_hpv": DatasetSpec(
            dataset_id="xena_tcga_hnsc_hpv",
            display_name="TCGA-HNSC Head & Neck (Xena, HPV binary)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_hnsc_hpv.npz",
                "synthetic_profile": {
                    "n_samples": 114,
                    "n_features": 20530,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_skcm": DatasetSpec(
            dataset_id="xena_tcga_skcm",
            display_name="TCGA-SKCM Melanoma (Xena, primary vs metastatic)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_skcm.npz",
                "synthetic_profile": {
                    "n_samples": 472,
                    "n_features": 20530,
                    "n_classes": 2,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_gbm": DatasetSpec(
            dataset_id="xena_tcga_gbm",
            display_name="TCGA-GBM Glioblastoma (Xena, 4-subtype)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_gbm.npz",
                "synthetic_profile": {
                    "n_samples": 164,
                    "n_features": 20530,
                    "n_classes": 4,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_stad": DatasetSpec(
            dataset_id="xena_tcga_stad",
            display_name="TCGA-STAD Stomach (Xena, histological 3-group)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_stad.npz",
                "synthetic_profile": {
                    "n_samples": 448,
                    "n_features": 20530,
                    "n_classes": 3,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_lihc": DatasetSpec(
            dataset_id="xena_tcga_lihc",
            display_name="TCGA-LIHC Liver (Xena, histologic grade 3-class)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_lihc.npz",
                "synthetic_profile": {
                    "n_samples": 415,
                    "n_features": 20530,
                    "n_classes": 3,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_ov": DatasetSpec(
            dataset_id="xena_tcga_ov",
            display_name="TCGA-OV Ovarian (Xena, histologic grade G2 vs G3)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_ov.npz",
                "synthetic_profile": {
                    "n_samples": 299,
                    "n_features": 20530,
                    "n_classes": 2,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_prad": DatasetSpec(
            dataset_id="xena_tcga_prad",
            display_name="TCGA-PRAD Prostate (Xena, Gleason score)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_prad.npz",
                "synthetic_profile": {
                    "n_samples": 550,
                    "n_features": 20530,
                    "n_classes": 5,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "xena_tcga_coad_cms": DatasetSpec(
            dataset_id="xena_tcga_coad_cms",
            display_name="TCGA-COAD Colorectal (Xena, histological 2-class)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/extended/xena_tcga_coad_cms.npz",
                "synthetic_profile": {
                    "n_samples": 323,
                    "n_features": 20530,
                    "n_classes": 2,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),

        # ── Results-validation datasets (rv_ prefix) ────────────────────────
        # Independent hold-out catalog for final results validation.
        # Not used in development tuning; see VALIDATION_RESULTS.md.

        # Group A: HDLSS small-n (scikit-feature)
        "rv_warpar10p": DatasetSpec(
            dataset_id="rv_warpar10p",
            display_name="warpAR10P (Scikit-feature, Face)",
            pipeline="fs",
            tier="medium",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/warpAR10P.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/warpAR10P.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 130,
                    "n_features": 2400,
                    "n_classes": 10,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "rv_yale": DatasetSpec(
            dataset_id="rv_yale",
            display_name="Yale Faces (Scikit-feature)",
            pipeline="fs",
            tier="hard",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/Yale.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/Yale.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 165,
                    "n_features": 1024,
                    "n_classes": 15,
                    "difficulty": "hard",
                },
                "n_final_features": 60,
            },
            extended_only=True,
        ),
        "rv_lung_discrete": DatasetSpec(
            dataset_id="rv_lung_discrete",
            display_name="Lung Discrete (Scikit-feature)",
            pipeline="fs",
            tier="hard",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/lung_discrete.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/lung_discrete.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 73,
                    "n_features": 325,
                    "n_classes": 7,
                    "difficulty": "hard",
                },
                "n_final_features": 50,
            },
            extended_only=True,
        ),

        # Group B: Moderate-scale datasets (scikit-feature)
        "rv_coil20": DatasetSpec(
            dataset_id="rv_coil20",
            display_name="COIL20 Objects (Scikit-feature)",
            pipeline="fs",
            tier="medium",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/COIL20.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/COIL20.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 1440,
                    "n_features": 1024,
                    "n_classes": 20,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "rv_relathe": DatasetSpec(
            dataset_id="rv_relathe",
            display_name="RELATHE Text (Scikit-feature)",
            pipeline="fs",
            tier="medium",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/RELATHE.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/RELATHE.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 1427,
                    "n_features": 4322,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "rv_pcmac": DatasetSpec(
            dataset_id="rv_pcmac",
            display_name="PCMAC Text (Scikit-feature)",
            pipeline="fs",
            tier="medium",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/PCMAC.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/PCMAC.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 1943,
                    "n_features": 3289,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "rv_isolet": DatasetSpec(
            dataset_id="rv_isolet",
            display_name="ISOLET Audio (Scikit-feature)",
            pipeline="fs",
            tier="hard",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/Isolet.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/Isolet.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 1560,
                    "n_features": 617,
                    "n_classes": 26,
                    "difficulty": "hard",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),
        "rv_basehock": DatasetSpec(
            dataset_id="rv_basehock",
            display_name="BASEHOCK Text (Scikit-feature)",
            pipeline="fs",
            tier="easy",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/BASEHOCK.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/BASEHOCK.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 1993,
                    "n_features": 4862,
                    "n_classes": 2,
                    "difficulty": "easy",
                },
                "n_final_features": 80,
            },
            extended_only=True,
        ),

        # Group C: Large-scale dataset (scikit-feature)
        "rv_usps": DatasetSpec(
            dataset_id="rv_usps",
            display_name="USPS Handwritten Digits (Scikit-feature)",
            pipeline="fs",
            tier="easy",
            loader_kind="mat_url_or_synth",
            params={
                "mat_url_options": [
                    {
                        "url": "https://jundongl.github.io/scikit-feature/OLD/datasets/USPS.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                    {
                        "url": "https://raw.githubusercontent.com/jundongl/scikit-feature/master/skfeature/data/USPS.mat",
                        "x_key": "X",
                        "y_key": "Y",
                    },
                ],
                "synthetic_profile": {
                    "n_samples": 9298,
                    "n_features": 256,
                    "n_classes": 10,
                    "difficulty": "easy",
                },
                "n_final_features": 60,
            },
            extended_only=True,
        ),

        # Group D: UCSC Xena / TCGA RNA-seq (manual NPZ)
        "rv_xena_tcga_thca": DatasetSpec(
            dataset_id="rv_xena_tcga_thca",
            display_name="TCGA-THCA Thyroid Cancer (Xena, 3-subtype)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/results_validation/xena_tcga_thca.npz",
                "synthetic_profile": {
                    "n_samples": 500,
                    "n_features": 20530,
                    "n_classes": 3,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "rv_xena_tcga_blca": DatasetSpec(
            dataset_id="rv_xena_tcga_blca",
            display_name="TCGA-BLCA Bladder Cancer (Xena, 5-subtype)",
            pipeline="fs",
            tier="hard",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/results_validation/xena_tcga_blca.npz",
                "synthetic_profile": {
                    "n_samples": 400,
                    "n_features": 20530,
                    "n_classes": 5,
                    "difficulty": "hard",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
        "rv_xena_tcga_cesc": DatasetSpec(
            dataset_id="rv_xena_tcga_cesc",
            display_name="TCGA-CESC Cervical Cancer (Xena, 2-class)",
            pipeline="fs",
            tier="medium",
            loader_kind="manual_or_synth",
            params={
                "default_local_path": "train_data/results_validation/xena_tcga_cesc.npz",
                "synthetic_profile": {
                    "n_samples": 300,
                    "n_features": 20530,
                    "n_classes": 2,
                    "difficulty": "medium",
                },
                "n_final_features": 100,
            },
            extended_only=True,
        ),
    }

    df_specs = {
        "df_synthetic_parametric": DatasetSpec(
            dataset_id="df_synthetic_parametric",
            display_name="Synthetic Controlled Parametric",
            pipeline="df",
            tier="easy",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "synthetic_parametric"},
        ),
        "df_actuarial_losses": DatasetSpec(
            dataset_id="df_actuarial_losses",
            display_name="Actuarial Loss Distributions",
            pipeline="df",
            tier="easy",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "actuarial"},
        ),
        "df_reliability_survival": DatasetSpec(
            dataset_id="df_reliability_survival",
            display_name="Reliability / Survival",
            pipeline="df",
            tier="easy",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "reliability"},
        ),
        "df_financial_returns": DatasetSpec(
            dataset_id="df_financial_returns",
            display_name="Financial Returns",
            pipeline="df",
            tier="medium",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "financial"},
        ),
        "df_hydrology_rainfall": DatasetSpec(
            dataset_id="df_hydrology_rainfall",
            display_name="Hydrology / Rainfall",
            pipeline="df",
            tier="medium",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "hydrology"},
        ),
        "df_internet_traffic": DatasetSpec(
            dataset_id="df_internet_traffic",
            display_name="Internet Traffic / File Sizes",
            pipeline="df",
            tier="medium",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "internet"},
        ),
        "df_contaminated_samples": DatasetSpec(
            dataset_id="df_contaminated_samples",
            display_name="Contaminated / Outlier-Heavy Samples",
            pipeline="df",
            tier="hard",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "contaminated"},
        ),
        "df_heaped_data": DatasetSpec(
            dataset_id="df_heaped_data",
            display_name="Heaped / Rounded Data",
            pipeline="df",
            tier="hard",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "heaped"},
        ),
        "df_tail_discrimination": DatasetSpec(
            dataset_id="df_tail_discrimination",
            display_name="Short vs Long Tail Discrimination",
            pipeline="df",
            tier="hard",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "tail_discrimination"},
        ),
        "df_near_symmetric_small_n": DatasetSpec(
            dataset_id="df_near_symmetric_small_n",
            display_name="Near-Symmetric Heavy-Tailed (Small n)",
            pipeline="df",
            tier="very_hard",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "near_symmetric"},
        ),
        "df_out_of_library": DatasetSpec(
            dataset_id="df_out_of_library",
            display_name="Misspecified / Out-of-Library",
            pipeline="df",
            tier="very_hard",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "out_of_library"},
        ),
        "df_multimodal_mixtures": DatasetSpec(
            dataset_id="df_multimodal_mixtures",
            display_name="Multi-Modal / Mixtures",
            pipeline="df",
            tier="very_hard",
            loader_kind="dist_benchmark",
            source_kind="dist_benchmark",
            params={"profile": "mixtures"},
        ),
    }

    integrated_specs = {
        "int_cdf_leukemia": DatasetSpec(
            dataset_id="int_cdf_leukemia",
            display_name="Integrated: Leukemia with CDF Transform",
            pipeline="integrated",
            tier="easy",
            loader_kind="integrated",
            params={"base_dataset": "leukemia_golub", "scenario": "cdf_transform"},
        ),
        "int_cdf_colon": DatasetSpec(
            dataset_id="int_cdf_colon",
            display_name="Integrated: Colon with CDF Transform",
            pipeline="integrated",
            tier="medium",
            loader_kind="integrated",
            params={"base_dataset": "colon_alon", "scenario": "cdf_transform"},
        ),
        "int_cdf_nci60": DatasetSpec(
            dataset_id="int_cdf_nci60",
            display_name="Integrated: NCI60 with CDF Transform",
            pipeline="integrated",
            tier="hard",
            loader_kind="integrated",
            params={"base_dataset": "nci60_ross", "scenario": "cdf_transform"},
        ),
        "int_low_gof_downweighting": DatasetSpec(
            dataset_id="int_low_gof_downweighting",
            display_name="Integrated: Low-GOF Downweighting",
            pipeline="integrated",
            tier="hard",
            loader_kind="integrated",
            params={"base_dataset": "colon_alon", "scenario": "low_gof_downweighting"},
        ),
        "int_distribution_stability_confidence": DatasetSpec(
            dataset_id="int_distribution_stability_confidence",
            display_name="Integrated: Distribution Stability Confidence",
            pipeline="integrated",
            tier="very_hard",
            loader_kind="integrated",
            params={"base_dataset": "nci60_ross", "scenario": "stability_signal"},
        ),
    }

    catalog: Dict[str, DatasetSpec] = {}
    catalog.update(fs_specs)
    catalog.update(df_specs)
    catalog.update(integrated_specs)
    return catalog

TIER_SOTA_DEFAULTS_STRICT: Dict[str, Tuple[float, float]] = {
    "easy": (0.90, 1.00),
    "medium": (0.75, 0.89),
    "hard": (0.50, 0.74),
    "very_hard": (0.00, 0.49),
}

TIER_SOTA_DEFAULTS_INFLATED: Dict[str, Tuple[float, float]] = {
    "easy": (0.95, 1.00),
    "medium": (0.85, 0.97),
    "hard": (0.65, 0.90),
    "very_hard": (0.45, 0.75),
}

# strict_holdout vs inflated protocol families (source: VALIDATION.md)
KNOWN_SOTA_PROTOCOL_RANGES: Dict[str, Dict[str, Tuple[float, float]]] = {
    "synthetic_easy_dfshift": {"strict_holdout": (0.90, 1.00), "inflated": (0.95, 1.00)},
    "synthetic_medium_mixed": {"strict_holdout": (0.72, 0.88), "inflated": (0.80, 0.93)},
    "synthetic_very_hard_sparse": {"strict_holdout": (0.30, 0.55), "inflated": (0.40, 0.65)},
    "leukemia_golub": {"strict_holdout": (0.93, 1.00), "inflated": (0.98, 1.00)},
    "dlbcl_shipp": {"strict_holdout": (0.90, 0.98), "inflated": (0.95, 1.00)},
    "colon_alon": {"strict_holdout": (0.80, 0.93), "inflated": (0.88, 0.96)},
    "gli_85": {"strict_holdout": (0.75, 0.87), "inflated": (0.80, 0.92)},
    "cns_pomeroy": {"strict_holdout": (0.74, 0.86), "inflated": (0.80, 0.90)},
    "breast_vantveer": {"strict_holdout": (0.75, 0.87), "inflated": (0.80, 0.90)},
    "ovarian_petricoin": {"strict_holdout": (0.95, 1.00), "inflated": (0.99, 1.00)},
    "srbct_khan": {"strict_holdout": (0.92, 1.00), "inflated": (0.97, 1.00)},
    "prostate_singh": {"strict_holdout": (0.88, 0.95), "inflated": (0.92, 1.00)},
    "smk_can_187": {"strict_holdout": (0.65, 0.80), "inflated": (0.75, 0.90)},
    "mll_microarray": {"strict_holdout": (0.88, 0.96), "inflated": (0.95, 1.00)},
    "lymphoma_3": {"strict_holdout": (0.80, 0.90), "inflated": (0.87, 0.97)},
    "lung_gordon": {"strict_holdout": (0.75, 0.88), "inflated": (0.82, 0.95)},
    "gcm_ramaswamy": {"strict_holdout": (0.50, 0.65), "inflated": (0.60, 0.78)},
    "tumor11_su": {"strict_holdout": (0.55, 0.80), "inflated": (0.65, 0.98)},
    "lymphoma_9": {"strict_holdout": (0.50, 0.70), "inflated": (0.65, 0.85)},
    "lymphoma_11": {"strict_holdout": (0.45, 0.65), "inflated": (0.60, 0.80)},
    "hf_breast_ge_mubashir1837": {"strict_holdout": (0.50, 0.70), "inflated": (0.75, 0.89)},
    "cumida_brain_gse50161": {"strict_holdout": (0.70, 0.88), "inflated": (0.82, 0.96)},
    "cumida_breast_gse45827": {"strict_holdout": (0.55, 0.72), "inflated": (0.70, 0.90)},
    "cumida_leukemia_subtypes": {"strict_holdout": (0.00, 0.49), "inflated": (0.70, 1.00)},
    "nci60_ross": {"strict_holdout": (0.40, 0.55), "inflated": (0.55, 0.75)},
    "nci60_strict_holdout": {"strict_holdout": (0.30, 0.50), "inflated": (0.55, 0.78)},
    "tumor9_openml": {"strict_holdout": (0.40, 0.62), "inflated": (0.50, 0.78)},
    "cll_sub_111": {"strict_holdout": (0.88, 0.95), "inflated": (0.95, 0.98)},
    "tox_171": {"strict_holdout": (0.84, 0.93), "inflated": (0.93, 0.98)},
    "gla_bra_180": {"strict_holdout": (0.70, 0.82), "inflated": (0.82, 0.90)},
    "carcinom_11class": {"strict_holdout": (0.72, 0.88), "inflated": (0.90, 0.98)},
    "glioma_50_4class": {"strict_holdout": (0.88, 0.97), "inflated": (0.96, 1.00)},
    "brain_tumor_2_50_4class": {"strict_holdout": (0.86, 0.95), "inflated": (0.95, 1.00)},
    "leukemia_1_72_3class": {"strict_holdout": (0.85, 0.96), "inflated": (0.95, 1.00)},
    "nci9_60_9class": {"strict_holdout": (0.35, 0.52), "inflated": (0.50, 0.62)},
    "nci_61_8class": {"strict_holdout": (0.50, 0.68), "inflated": (0.65, 0.75)},
    "orlraws10p": {"strict_holdout": (0.95, 1.00), "inflated": (0.98, 1.00)},
    "warp_pie10p": {"strict_holdout": (0.95, 1.00), "inflated": (0.98, 1.00)},
    "pixraw10p": {"strict_holdout": (0.96, 1.00), "inflated": (0.99, 1.00)},
    "arcene_nips03": {"strict_holdout": (0.75, 0.89), "inflated": (0.85, 0.97)},
    "madelon_nips03": {"strict_holdout": (0.75, 0.89), "inflated": (0.85, 0.97)},
    "gisette_nips03": {"strict_holdout": (0.75, 0.89), "inflated": (0.85, 0.97)},
    "dexter_nips03": {"strict_holdout": (0.75, 0.89), "inflated": (0.85, 0.97)},
    "dorothea_nips03": {"strict_holdout": (0.50, 0.74), "inflated": (0.65, 0.90)},
    # Extended-only CuMiDa datasets
    "cumida_prostate_gse6919": {"strict_holdout": (0.80, 0.92), "inflated": (0.88, 0.97)},
    "cumida_ovarian_gse26712": {"strict_holdout": (0.80, 0.92), "inflated": (0.88, 0.97)},
    "cumida_lung_gse19804": {"strict_holdout": (0.85, 0.95), "inflated": (0.92, 0.99)},
    "cumida_colorectal_gse44861": {"strict_holdout": (0.78, 0.90), "inflated": (0.85, 0.95)},
    "cumida_gastric_gse54129": {"strict_holdout": (0.90, 0.98), "inflated": (0.95, 1.00)},
    "cumida_pancreatic_gse16515": {"strict_holdout": (0.80, 0.92), "inflated": (0.88, 0.97)},
    "cumida_renal_gse53757": {"strict_holdout": (0.80, 0.92), "inflated": (0.88, 0.97)},
    "cumida_headneck_gse12452": {"strict_holdout": (0.78, 0.90), "inflated": (0.85, 0.95)},
    # Extended-only UCSC Xena / TCGA datasets
    "xena_tcga_brca": {"strict_holdout": (0.55, 0.72), "inflated": (0.65, 0.85)},
    "xena_tcga_luad": {"strict_holdout": (0.55, 0.70), "inflated": (0.65, 0.82)},
    "xena_tcga_ucec": {"strict_holdout": (0.50, 0.68), "inflated": (0.60, 0.80)},
    "xena_tcga_lgg": {"strict_holdout": (0.55, 0.72), "inflated": (0.65, 0.85)},
    "xena_tcga_kirc": {"strict_holdout": (0.50, 0.68), "inflated": (0.60, 0.80)},
    # Extended-only UCSC Xena / TCGA expansion (Phase 2)
    "xena_tcga_hnsc_hpv": {"strict_holdout": (0.82, 0.93), "inflated": (0.90, 0.99)},
    "xena_tcga_skcm": {"strict_holdout": (0.60, 0.78), "inflated": (0.75, 0.92)},
    "xena_tcga_gbm": {"strict_holdout": (0.55, 0.72), "inflated": (0.70, 0.90)},
    "xena_tcga_stad": {"strict_holdout": (0.58, 0.75), "inflated": (0.72, 0.90)},
    "xena_tcga_lihc": {"strict_holdout": (0.52, 0.68), "inflated": (0.65, 0.82)},
    "xena_tcga_ov": {"strict_holdout": (0.48, 0.65), "inflated": (0.62, 0.80)},
    "xena_tcga_prad": {"strict_holdout": (0.45, 0.65), "inflated": (0.60, 0.80)},
    "xena_tcga_coad_cms": {"strict_holdout": (0.58, 0.75), "inflated": (0.72, 0.90)},
    # Results-validation datasets (rv_ prefix) — VALIDATION_RESULTS.md
    "rv_warpar10p": {"strict_holdout": (0.70, 0.85), "inflated": (0.80, 0.92)},
    "rv_yale": {"strict_holdout": (0.55, 0.72), "inflated": (0.65, 0.85)},
    "rv_lung_discrete": {"strict_holdout": (0.60, 0.78), "inflated": (0.72, 0.90)},
    "rv_coil20": {"strict_holdout": (0.85, 0.95), "inflated": (0.92, 0.99)},
    "rv_relathe": {"strict_holdout": (0.82, 0.92), "inflated": (0.88, 0.96)},
    "rv_pcmac": {"strict_holdout": (0.80, 0.90), "inflated": (0.86, 0.95)},
    "rv_isolet": {"strict_holdout": (0.60, 0.78), "inflated": (0.72, 0.88)},
    "rv_basehock": {"strict_holdout": (0.90, 0.97), "inflated": (0.95, 0.99)},
    "rv_usps": {"strict_holdout": (0.90, 0.97), "inflated": (0.95, 0.99)},
    "rv_xena_tcga_thca": {"strict_holdout": (0.55, 0.72), "inflated": (0.65, 0.85)},
    "rv_xena_tcga_blca": {"strict_holdout": (0.50, 0.68), "inflated": (0.60, 0.82)},
    "rv_xena_tcga_cesc": {"strict_holdout": (0.75, 0.88), "inflated": (0.82, 0.95)},
}

def sota_ranges_for_dataset(dataset_id: str, tier: str) -> Tuple[Tuple[float, float], Tuple[float, float]]:
    ds_id = str(dataset_id)
    tier_norm = str(tier).strip().lower()
    ranges = KNOWN_SOTA_PROTOCOL_RANGES.get(ds_id)
    if ranges is None:
        strict = TIER_SOTA_DEFAULTS_STRICT.get(tier_norm, (0.0, 1.0))
        inflated = TIER_SOTA_DEFAULTS_INFLATED.get(tier_norm, strict)
        return strict, inflated
    strict = tuple(ranges.get("strict_holdout") or TIER_SOTA_DEFAULTS_STRICT.get(tier_norm, (0.0, 1.0)))
    inflated = tuple(ranges.get("inflated") or TIER_SOTA_DEFAULTS_INFLATED.get(tier_norm, strict))
    return (float(strict[0]), float(strict[1])), (float(inflated[0]), float(inflated[1]))

def infer_domain(spec: DatasetSpec) -> str:
    ds_id = str(spec.dataset_id).strip().lower()
    kind = str(spec.loader_kind)
    if str(spec.source_kind) == "synthetic":
        return "synthetic"
    if kind == "face_proxy_or_synth":
        return "non_genomic"
    if ds_id.endswith("_nips03"):
        return "non_genomic"
    explicit = str((spec.params or {}).get("domain", "")).strip().lower()
    if explicit in {"genomics", "non_genomic", "synthetic"}:
        return explicit
    return "genomics"


def infer_platform(spec: DatasetSpec) -> str:
    ds_id = str(spec.dataset_id).strip().lower()
    params = dict(spec.params or {})

    explicit = str(params.get("platform", "") or "").strip()
    if explicit:
        return explicit

    if str(spec.source_kind) == "synthetic":
        return "synthetic"
    if ds_id == "ovarian_petricoin":
        return "mass-spec"

    affy_hgu95 = {
        "leukemia_golub",
        "dlbcl_shipp",
        "prostate_singh",
        "mll_microarray",
        "nci60_ross",
        "nci60_strict_holdout",
    }
    if ds_id in affy_hgu95:
        return "Affy HG-U95"

    affy_hgu133a = {
        "colon_alon",
        "gli_85",
        "smk_can_187",
        "lymphoma_3",
        "lung_gordon",
        "gcm_ramaswamy",
        "tumor11_su",
        "lymphoma_9",
        "lymphoma_11",
    }
    if ds_id in affy_hgu133a:
        return "Affy HG-U133A"

    return "cDNA"

def infer_provenance(spec: DatasetSpec) -> str:
    kind = str(spec.loader_kind)
    params = dict(spec.params or {})
    if kind == "openml_or_synth":
        opts = list(params.get("openml_options", []) or [])
        if opts:
            return f"openml:{opts[0]}"
        return "openml"
    if kind in {"mat_url_or_synth", "tab_url_or_synth"}:
        key = "mat_url_options" if kind == "mat_url_or_synth" else "tab_url_options"
        opts = list(params.get(key, []) or [])
        if opts and isinstance(opts[0], dict) and opts[0].get("url"):
            return str(opts[0].get("url"))
        return kind
    if kind in {"nci60_url_or_synth", "nci60_proxy_or_synth"}:
        return kind
    if kind == "face_proxy_or_synth":
        return "sklearn:olivetti_faces"
    if kind == "manual_or_synth":
        env_name = params.get("local_path_env")
        if env_name:
            return f"env:{env_name}"
        default_path = params.get("default_local_path")
        if default_path:
            return f"local:{default_path}"
        return "manual"
    if kind == "integrated":
        base = params.get("base_dataset")
        return f"integrated_base:{base}"
    if kind == "dist_benchmark":
        return "dist_benchmark"
    return kind

def _finalize_spec(spec: DatasetSpec) -> DatasetSpec:
    prov = infer_provenance(spec)
    dom = infer_domain(spec)
    platform = infer_platform(spec)
    out = replace(spec, provenance=str(prov), domain=str(dom), platform=str(platform))
    if out.pipeline in {"fs", "integrated"} and out.source_kind in {"synthetic", "validation_catalog"}:
        sota_id = out.dataset_id
        if out.pipeline == "integrated":
            # Keep benchmark behavior stable: integrated datasets inherit their parent
            # dataset's SOTA ranges when available (per VALIDATION.md / benchmark runner).
            base = None
            try:
                base = out.params.get("base_dataset")
            except Exception as exc:
                base = None
            base = str(base).strip() if base else ""
            if base and base in KNOWN_SOTA_PROTOCOL_RANGES:
                sota_id = base

        strict, inflated = sota_ranges_for_dataset(sota_id, out.tier)
        out = replace(out, sota_holdout_bal_acc=strict, sota_inflated_bal_acc=inflated)
    return out

def _build_registry() -> Dict[str, DatasetSpec]:
    reg = _build_dataset_registry()
    # Synthetic benchmark datasets (from run_df_fs_sota_benchmark.py)
    reg.update({
        "synthetic_easy_dfshift": DatasetSpec(
            dataset_id="synthetic_easy_dfshift",
            display_name="Synthetic Easy DF-Shift",
            pipeline="fs",
            tier="easy",
            loader_kind="synthetic_benchmark",
            params={
                "fs_fraction": 0.40,
                "n_final_features": 40,
                "notes": "High-separation synthetic with class-conditional marginal shifts.",
            },
            source_kind="synthetic",
        ),
        "synthetic_medium_mixed": DatasetSpec(
            dataset_id="synthetic_medium_mixed",
            display_name="Synthetic Medium Mixed",
            pipeline="fs",
            tier="medium",
            loader_kind="synthetic_benchmark",
            params={
                "fs_fraction": 0.40,
                "n_final_features": 50,
                "notes": "Moderate overlap with skew/heavy-tail perturbations.",
            },
            source_kind="synthetic",
        ),
        "synthetic_very_hard_sparse": DatasetSpec(
            dataset_id="synthetic_very_hard_sparse",
            display_name="Synthetic Very Hard Sparse",
            pipeline="fs",
            tier="very_hard",
            loader_kind="synthetic_benchmark",
            params={
                "fs_fraction": 0.40,
                "n_final_features": 60,
                "notes": "Multiclass sparse-signal problem with contamination/heaping.",
            },
            source_kind="synthetic",
        ),
    })
    return {k: _finalize_spec(v) for k, v in reg.items()}

DATASET_REGISTRY: Dict[str, DatasetSpec] = _build_registry()



import numpy as np

from tabnetics.domains.base import resolve_dataset_catalog_context
from tabnetics.domains.bio import (
    apply_multiomics_adapter_train_test,
    infer_prefilter_data_domain,
)
from tabnetics.domains.face import apply_face_domain_projection


def _face_like_data(seed: int = 0) -> tuple[np.ndarray, np.ndarray]:
    rng = np.random.default_rng(seed)
    y = np.repeat(np.arange(3), 8)
    X = rng.normal(size=(24, 18))
    X[y == 0, :4] += 2.0
    X[y == 1, 4:8] += 2.0
    X[y == 2, 8:12] += 2.0
    return X, y


def _omics_like_train_test(seed: int = 0) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    rng = np.random.default_rng(seed)
    y = np.tile(np.array([0, 1, 2]), 6)
    X = rng.normal(size=(18, 8))
    X[y == 0, :2] += 2.5
    X[y == 1, 2:4] += 2.5
    X[y == 2, 4:6] += 2.5
    return X[:12], y[:12], X[12:]


def test_resolve_dataset_catalog_context_marks_face_domain() -> None:
    context = resolve_dataset_catalog_context("orlraws10p")

    assert context.dataset_id == "orlraws10p"
    assert context.found_in_catalog is True
    assert context.is_face_domain is True


def test_infer_prefilter_data_domain_detects_rnaseq() -> None:
    assert infer_prefilter_data_domain("RNA-seq counts") == "rnaseq"
    assert infer_prefilter_data_domain("Affy HG-U133A") == "auto"


def test_apply_face_domain_projection_extracts_face_specific_state() -> None:
    X, y = _face_like_data(11)
    context = resolve_dataset_catalog_context("orlraws10p")

    result = apply_face_domain_projection(
        X_train_imp=X[:18],
        y_train=y[:18],
        X_test_imp=X[18:],
        enabled=True,
        dataset_name="orlraws10p",
        dataset_context=context,
        seed=11,
    )

    assert result.meta["face_projection_applied"] is True
    assert result.meta["face_projection_is_face_domain"] is True
    assert result.state.pca_model is not None
    assert result.X_train.shape[1] >= 1
    assert np.all(np.isfinite(result.X_train))
    assert np.all(np.isfinite(result.X_test))


def test_apply_multiomics_adapter_train_test_appends_latent_features() -> None:
    X_train, y_train, X_test = _omics_like_train_test(7)

    X_train_out, X_test_out, meta = apply_multiomics_adapter_train_test(
        X_train=X_train,
        y_train=y_train,
        X_test=X_test,
        adapter_mode="split_halves",
        integrator="mb_plsda",
        n_components=2,
    )

    assert meta["multiomics_adapter_applied"] is True
    assert int(meta["multiomics_latent_dim"]) >= 1
    assert X_train_out.shape[1] > X_train.shape[1]
    assert X_test_out.shape[1] == X_train_out.shape[1]
    assert np.all(np.isfinite(X_train_out))
    assert np.all(np.isfinite(X_test_out))



"""Wrapper feature selection methods (iterative pruning)."""

import numpy as np
from sklearn.covariance import LedoitWolf
from time import perf_counter


def iterative_pruning_cpss_overlay(
    X_pool, y_arr, target, base_selected, base_relevance, base_score, *,
    use_cpss_overlay,
    cpss_pairs,
    cpss_stability_threshold,
    cpss_min_stable_features,
    cpss_min_jaccard,
    cpss_max_score_drop,
    random_state,
    mi_scorer,
    f_scorer,
    normalize_fn,
    wrapper_score_fn,
):
    """
    A16: CPSS-style complementary-pairs stability overlay for bounded pruning.
    """
    n_samples, n_pool = X_pool.shape
    base_selected = np.asarray(base_selected, dtype=int).ravel()
    out = {
        "iterative_pruning_cpss_overlay_enabled": bool(use_cpss_overlay),
        "iterative_pruning_cpss_switch_applied": False,
        "iterative_pruning_cpss_switch_reason": "disabled",
        "iterative_pruning_cpss_pairs": int(cpss_pairs),
        "iterative_pruning_cpss_subruns": 0,
        "iterative_pruning_cpss_stability_threshold": float(cpss_stability_threshold),
        "iterative_pruning_cpss_min_stable_features": int(cpss_min_stable_features),
        "iterative_pruning_cpss_min_jaccard": float(cpss_min_jaccard),
        "iterative_pruning_cpss_max_score_drop": float(cpss_max_score_drop),
        "iterative_pruning_cpss_stable_feature_count": 0,
        "iterative_pruning_cpss_overlap_jaccard": float("nan"),
        "iterative_pruning_cpss_overlap_recall": float("nan"),
        "iterative_pruning_cpss_base_score": float(base_score) if np.isfinite(base_score) else float("nan"),
        "iterative_pruning_cpss_overlay_score": float("nan"),
        "iterative_pruning_cpss_overlay_score_evaluated": False,
        "iterative_pruning_cpss_selected_local": np.asarray(base_selected, dtype=int),
        "iterative_pruning_cpss_freq_q50": float("nan"),
        "iterative_pruning_cpss_freq_q90": float("nan"),
    }

    if not use_cpss_overlay:
        return out
    if n_pool <= 0:
        out["iterative_pruning_cpss_switch_reason"] = "empty_pool"
        return out
    if n_samples < 8:
        out["iterative_pruning_cpss_switch_reason"] = "insufficient_samples"
        return out
    if np.unique(y_arr).size < 2:
        out["iterative_pruning_cpss_switch_reason"] = "single_class"
        return out

    rng = np.random.default_rng(random_state + 1337)
    support_counts = np.zeros(n_pool, dtype=float)
    n_subruns = 0
    target = int(max(1, min(target, n_pool)))

    for pair_idx in range(int(cpss_pairs)):
        perm = rng.permutation(n_samples)
        half = int(perm.size // 2)
        if half < 4:
            continue
        subsets = (perm[:half], perm[half: 2 * half])
        for subset_idx in subsets:
            subset_idx = np.asarray(subset_idx, dtype=int)
            if subset_idx.size < 4:
                continue
            y_sub = np.asarray(y_arr[subset_idx])
            if np.unique(y_sub).size < 2:
                continue
            X_sub = np.asarray(X_pool[subset_idx], dtype=float)
            try:
                mi_scores = np.asarray(mi_scorer(X_sub, y_sub, random_state=random_state), dtype=float).ravel()
            except Exception as exc:
                mi_scores = np.zeros(n_pool, dtype=float)
            mi_scores = np.nan_to_num(mi_scores, nan=0.0, posinf=0.0, neginf=0.0)
            try:
                f_scores, _ = f_scorer(X_sub, y_sub)
                f_scores = np.asarray(f_scores, dtype=float).ravel()
            except Exception as exc:
                f_scores = np.zeros(n_pool, dtype=float)
            f_scores = np.nan_to_num(f_scores, nan=0.0, posinf=0.0, neginf=0.0)
            local = 0.60 * normalize_fn(mi_scores) + 0.40 * normalize_fn(f_scores)
            local = np.asarray(normalize_fn(local), dtype=float).ravel()
            if local.size != n_pool:
                continue
            chosen = np.argsort(local)[::-1][:target]
            support_counts[np.asarray(chosen, dtype=int)] += 1.0
            n_subruns += 1

    out["iterative_pruning_cpss_subruns"] = int(n_subruns)
    if n_subruns <= 0:
        out["iterative_pruning_cpss_switch_reason"] = "no_valid_subruns"
        return out

    freq = support_counts / float(n_subruns)
    stable_idx = np.where(freq >= float(cpss_stability_threshold))[0]
    out["iterative_pruning_cpss_stable_feature_count"] = int(stable_idx.size)
    out["iterative_pruning_cpss_freq_q50"] = float(np.quantile(freq, 0.50))
    out["iterative_pruning_cpss_freq_q90"] = float(np.quantile(freq, 0.90))

    if stable_idx.size < int(cpss_min_stable_features):
        out["iterative_pruning_cpss_switch_reason"] = "insufficient_stable_support"
        return out

    base_norm = normalize_fn(base_relevance)
    stable_boost = np.zeros(n_pool, dtype=float)
    stable_boost[stable_idx] = 1.0
    combined = 0.55 * freq + 0.25 * base_norm + 0.20 * stable_boost
    combined = np.asarray(normalize_fn(combined), dtype=float).ravel()
    candidate = np.argsort(combined)[::-1][:target]
    candidate = np.asarray(candidate, dtype=int)
    if candidate.size == 0:
        out["iterative_pruning_cpss_switch_reason"] = "no_candidate"
        return out

    base_set = set(int(i) for i in np.asarray(base_selected, dtype=int).tolist())
    cand_set = set(int(i) for i in candidate.tolist())
    union = len(base_set.union(cand_set))
    inter = len(base_set.intersection(cand_set))
    jaccard = float(inter / union) if union > 0 else 1.0
    recall = float(inter / max(1, len(base_set)))
    out["iterative_pruning_cpss_overlap_jaccard"] = float(jaccard)
    out["iterative_pruning_cpss_overlap_recall"] = float(recall)

    if jaccard < float(cpss_min_jaccard):
        out["iterative_pruning_cpss_switch_reason"] = "overlap_below_threshold"
        return out

    overlay_score = wrapper_score_fn(X_pool, y_arr, candidate)
    out["iterative_pruning_cpss_overlay_score_evaluated"] = True
    if np.isfinite(overlay_score):
        out["iterative_pruning_cpss_overlay_score"] = float(overlay_score)
    else:
        out["iterative_pruning_cpss_switch_reason"] = "overlay_score_not_finite"
        return out

    if np.isfinite(base_score):
        if float(overlay_score) < float(base_score) - float(cpss_max_score_drop):
            out["iterative_pruning_cpss_switch_reason"] = "score_drop_exceeds_threshold"
            return out

    out["iterative_pruning_cpss_switch_applied"] = True
    out["iterative_pruning_cpss_switch_reason"] = "switch_applied"
    out["iterative_pruning_cpss_selected_local"] = np.asarray(candidate, dtype=int)
    return out


def iterative_redundancy_pruning_core(
    X, y, n_target_features, runtime_bounded=False, *,
    # Callables
    prefilter_fn,
    pareto_prefilter_fn,
    normalize_fn,
    wrapper_score_fn,
    cpss_overlay_fn,
    mi_scorer,
    f_scorer,
    # Data
    score_cache,
    problem_type,
    random_state,
    mrmr_max_features,
    # Pruning config
    iterative_pruning_pool_factor,
    iterative_pruning_redundancy_weight,
    iterative_pruning_max_rounds,
    iterative_pruning_min_improvement,
    iterative_pruning_max_cumulative_loss,
    # Pareto prefilter config
    iterative_pruning_class_pareto_prefilter_enabled,
    iterative_pruning_class_pareto_top_per_class,
    iterative_pruning_class_pareto_global_fraction,
    iterative_pruning_class_pareto_minority_boost,
    # Bounded config
    iterative_pruning_bounded_prefilter_cap,
    iterative_pruning_bounded_max_evaluations,
    iterative_pruning_bounded_max_runtime_seconds,
    iterative_pruning_bounded_candidate_fraction,
    iterative_pruning_bounded_min_candidates,
    iterative_pruning_bounded_enable_class_gating,
    iterative_pruning_bounded_multiclass_scale,
    iterative_pruning_bounded_imbalance_trigger,
    iterative_pruning_bounded_imbalance_scale,
    # CPSS config
    iterative_pruning_bounded_use_cpss_overlay,
    iterative_pruning_bounded_cpss_pairs,
    iterative_pruning_bounded_cpss_stability_threshold,
    iterative_pruning_bounded_cpss_min_stable_features,
    iterative_pruning_bounded_cpss_min_jaccard,
    iterative_pruning_bounded_cpss_max_score_drop,
):
    """
    ieGENES-style iterative wrapper pruning core.
    When ``runtime_bounded=True`` (A15), candidate evaluations are explicitly
    budgeted by prefilter cap, per-round candidate budget, max evaluations,
    max runtime, and class-aware candidate gating.
    """
    n_samples, n_features = X.shape
    if n_features == 0:
        return {}, {}

    y_arr = np.asarray(y)
    classes, counts = np.unique(y_arr, return_counts=True)
    n_classes = int(classes.size)
    min_class_count = int(np.min(counts)) if counts.size else 0
    imbalance_ratio = float(np.max(counts) / max(1, min_class_count)) if counts.size else 1.0

    pool_cap = int(
        min(
            n_features,
            max(mrmr_max_features, min(640, max(96, 8 * int(max(1, n_target_features))))),
        )
    )
    if runtime_bounded:
        pool_cap = int(min(pool_cap, int(iterative_pruning_bounded_prefilter_cap)))

    method_name = "iterative_redundancy_pruning_bounded" if runtime_bounded else "iterative_redundancy_pruning"

    if problem_type == "classification" and iterative_pruning_class_pareto_prefilter_enabled:
        pool_idx, pareto_meta = pareto_prefilter_fn(X, y_arr, max_features=pool_cap)
    else:
        pool_idx = prefilter_fn(X, y_arr, max_features=pool_cap)
        pareto_meta = {
            "iterative_pruning_pareto_prefilter_enabled": bool(iterative_pruning_class_pareto_prefilter_enabled),
            "iterative_pruning_pareto_prefilter_applied": False,
            "iterative_pruning_pareto_prefilter_reason": (
                "non_classification_problem" if problem_type != "classification" else "disabled"
            ),
            "iterative_pruning_pareto_prefilter_candidate_universe": int(pool_idx.size),
            "iterative_pruning_pareto_prefilter_front_size": 0,
            "iterative_pruning_pareto_prefilter_classes_used": 0,
            "iterative_pruning_pareto_prefilter_top_per_class": int(iterative_pruning_class_pareto_top_per_class),
            "iterative_pruning_pareto_prefilter_global_fraction": float(
                iterative_pruning_class_pareto_global_fraction
            ),
            "iterative_pruning_pareto_prefilter_minority_boost": float(
                iterative_pruning_class_pareto_minority_boost
            ),
            "iterative_pruning_pareto_prefilter_class_top_hits": {},
        }

    if pool_idx.size == 0:
        return {}, {}

    X_pool = np.asarray(X[:, pool_idx], dtype=float)
    n_pool = int(X_pool.shape[1])
    target = int(min(max(1, n_target_features), n_pool))

    if n_pool <= target:
        scores = np.zeros(n_features, dtype=float)
        scores[pool_idx] = 1.0
        selected = np.asarray(pool_idx[:target], dtype=int)
        results = {
            "selected_indices": selected,
            "scores": {int(idx): float(scores[int(idx)]) for idx in selected},
            "method": method_name,
            "iterative_pruning_initial_size": int(n_pool),
            "iterative_pruning_final_size": int(selected.size),
            "iterative_pruning_rounds": 0,
            "iterative_pruning_pool_size": int(n_pool),
            "iterative_pruning_runtime_bounded": bool(runtime_bounded),
            "iterative_pruning_stop_reason": "pool_not_larger_than_target",
            "iterative_pruning_evaluations": 0,
        }
        results.update(pareto_meta)
        return results, {i: float(scores[i]) for i in range(n_features)}

    # Use score cache if available (P1-1: eliminates redundant MI/F-test
    # computations — scores already computed on full X_uncorr in fit_transform).
    # pool_idx indexes into the same X_uncorr columns the cache was built on.
    if score_cache is not None:
        mi = np.asarray(score_cache.mi_scores[pool_idx], dtype=float).ravel()
        mi = np.nan_to_num(mi, nan=0.0, posinf=0.0, neginf=0.0)
        f_vals = np.asarray(score_cache.f_scores[0][pool_idx], dtype=float).ravel()
        f_vals = np.nan_to_num(f_vals, nan=0.0, posinf=0.0, neginf=0.0)
    else:
        try:
            mi = np.asarray(mi_scorer(X_pool, y_arr, random_state=random_state), dtype=float).ravel()
        except Exception as exc:
            mi = np.zeros(n_pool, dtype=float)
        mi = np.nan_to_num(mi, nan=0.0, posinf=0.0, neginf=0.0)
        try:
            f_vals, _ = f_scorer(X_pool, y_arr)
            f_vals = np.asarray(f_vals, dtype=float).ravel()
        except Exception as exc:
            f_vals = np.zeros(n_pool, dtype=float)
        f_vals = np.nan_to_num(f_vals, nan=0.0, posinf=0.0, neginf=0.0)
    base_relevance = 0.60 * normalize_fn(mi) + 0.40 * normalize_fn(f_vals)
    base_relevance = np.asarray(base_relevance, dtype=float).ravel()
    base_relevance = np.nan_to_num(base_relevance, nan=0.0, posinf=0.0, neginf=0.0)

    init_size = int(
        min(
            n_pool,
            max(
                target + 2,
                int(round(iterative_pruning_pool_factor * float(target))),
            ),
        )
    )
    current = np.asarray(np.argsort(base_relevance)[::-1][:init_size], dtype=int)
    if current.size <= target:
        current = np.asarray(np.argsort(base_relevance)[::-1][:target], dtype=int)

    evaluations = 0
    removed_features = []
    removed_deltas = []
    cumulative_delta = 0.0
    rounds = 0
    stop_reason = "max_rounds_reached"
    start_t = perf_counter()
    per_round_candidate_budget = []

    max_evaluations = int(iterative_pruning_bounded_max_evaluations) if runtime_bounded else 0
    max_runtime_seconds = (
        float(iterative_pruning_bounded_max_runtime_seconds) if runtime_bounded else float("inf")
    )
    candidate_fraction = (
        float(iterative_pruning_bounded_candidate_fraction) if runtime_bounded else 1.0
    )
    min_candidates = int(iterative_pruning_bounded_min_candidates) if runtime_bounded else 1
    class_gating = bool(runtime_bounded and iterative_pruning_bounded_enable_class_gating)

    best_score = wrapper_score_fn(X_pool, y_arr, current)
    evaluations += 1
    if not np.isfinite(best_score):
        best_score = float("-inf")
        stop_reason = "initial_score_not_finite"

    corr_estimator = "corrcoef"
    corr_shrinkage = float("nan")
    if int(X_pool.shape[0]) < 2 or int(X_pool.shape[1]) < 2:
        corr_estimator = "degenerate"
        corr_pool = np.zeros((int(X_pool.shape[1]), int(X_pool.shape[1])), dtype=float)
    elif int(X_pool.shape[0]) < 2 * int(X_pool.shape[1]):
        corr_estimator = "ledoit_wolf"
        try:
            lw = LedoitWolf().fit(X_pool)
            cov = np.asarray(lw.covariance_, dtype=float)
            corr_shrinkage = float(getattr(lw, "shrinkage_", float("nan")))
            std = np.sqrt(np.maximum(0.0, np.diag(cov)))
            denom = np.outer(std, std)
            with np.errstate(invalid="ignore", divide="ignore"):
                corr = np.divide(cov, denom, out=np.zeros_like(cov), where=denom > 1e-12)
            corr_pool = np.abs(np.nan_to_num(corr, nan=0.0, posinf=0.0, neginf=0.0))
        except Exception as exc:
            corr_estimator = "corrcoef_fallback"
            with np.errstate(invalid="ignore"):
                corr = np.corrcoef(X_pool, rowvar=False)
            corr_pool = np.abs(np.nan_to_num(corr, nan=0.0, posinf=0.0, neginf=0.0))
    else:
        with np.errstate(invalid="ignore"):
            corr = np.corrcoef(X_pool, rowvar=False)
        corr_pool = np.abs(np.nan_to_num(corr, nan=0.0, posinf=0.0, neginf=0.0))
    np.fill_diagonal(corr_pool, 0.0)

    while current.size > target and rounds < int(iterative_pruning_max_rounds):
        if runtime_bounded and (perf_counter() - start_t) >= max_runtime_seconds:
            stop_reason = "runtime_budget_exhausted"
            break
        if runtime_bounded and max_evaluations > 0 and evaluations >= max_evaluations:
            stop_reason = "evaluation_budget_exhausted"
            break

        rounds += 1
        local_corr = corr_pool[np.ix_(current, current)]
        mean_corr = np.mean(local_corr, axis=1) if local_corr.size else np.zeros(current.size, dtype=float)
        rel_local = base_relevance[current]
        priority = (
            float(iterative_pruning_redundancy_weight) * normalize_fn(mean_corr)
            - (1.0 - float(iterative_pruning_redundancy_weight)) * normalize_fn(rel_local)
        )
        candidate_order = np.asarray(np.argsort(priority)[::-1], dtype=int)

        if runtime_bounded:
            budget = int(max(min_candidates, int(np.ceil(candidate_fraction * float(candidate_order.size)))))
            if class_gating and n_classes > 2:
                budget = int(max(1, int(np.ceil(budget * float(iterative_pruning_bounded_multiclass_scale)))))
            if class_gating and imbalance_ratio >= float(iterative_pruning_bounded_imbalance_trigger):
                budget = int(max(1, int(np.ceil(budget * float(iterative_pruning_bounded_imbalance_scale)))))
            budget = int(min(candidate_order.size, max(1, budget)))
            candidate_order = candidate_order[:budget]
            per_round_candidate_budget.append(int(budget))
        else:
            per_round_candidate_budget.append(int(candidate_order.size))

        best_remove_idx = None
        best_remove_score = float("-inf")
        for local_idx in candidate_order.tolist():
            if runtime_bounded and (perf_counter() - start_t) >= max_runtime_seconds:
                stop_reason = "runtime_budget_exhausted"
                break
            if runtime_bounded and max_evaluations > 0 and evaluations >= max_evaluations:
                stop_reason = "evaluation_budget_exhausted"
                break

            local_idx = int(local_idx)
            trial = np.delete(current, local_idx)
            trial_score = wrapper_score_fn(X_pool, y_arr, trial)
            evaluations += 1
            if not np.isfinite(trial_score):
                continue
            if trial_score > best_remove_score:
                best_remove_score = float(trial_score)
                best_remove_idx = int(local_idx)

        if stop_reason in {"runtime_budget_exhausted", "evaluation_budget_exhausted"}:
            break

        if best_remove_idx is None:
            stop_reason = "no_valid_removal_candidate"
            break

        delta = float(best_remove_score - best_score) if np.isfinite(best_score) else float("inf")
        if np.isfinite(delta) and delta < float(iterative_pruning_min_improvement):
            stop_reason = "min_improvement_not_met"
            break
        delta_eff = float(delta) if np.isfinite(delta) else 0.0
        if (
            np.isfinite(delta_eff)
            and float(iterative_pruning_max_cumulative_loss) > 0.0
            and (cumulative_delta + delta_eff) < -float(iterative_pruning_max_cumulative_loss)
        ):
            stop_reason = "cumulative_loss_budget_exhausted"
            break

        removed_feature_local = int(current[best_remove_idx])
        removed_features.append(int(pool_idx[removed_feature_local]))
        removed_deltas.append(float(delta_eff))
        cumulative_delta += float(delta_eff)
        current = np.delete(current, best_remove_idx)
        best_score = float(best_remove_score)

        if current.size <= target:
            stop_reason = "target_size_reached"
            break

    if current.size > target:
        corr_local = corr_pool[np.ix_(current, current)]
        mean_corr = np.mean(corr_local, axis=1) if corr_local.size else np.zeros(current.size, dtype=float)
        rel_local = base_relevance[current]
        ranking = np.argsort(
            0.65 * normalize_fn(rel_local)
            + 0.35 * (1.0 - normalize_fn(mean_corr))
        )[::-1]
        current = current[np.asarray(ranking[:target], dtype=int)]

    selected_local = np.asarray(current[:target], dtype=int)
    cpss_meta = {
        "iterative_pruning_cpss_overlay_enabled": bool(runtime_bounded and iterative_pruning_bounded_use_cpss_overlay),
        "iterative_pruning_cpss_switch_applied": False,
        "iterative_pruning_cpss_switch_reason": "disabled",
        "iterative_pruning_cpss_pairs": int(iterative_pruning_bounded_cpss_pairs),
        "iterative_pruning_cpss_subruns": 0,
        "iterative_pruning_cpss_stability_threshold": float(iterative_pruning_bounded_cpss_stability_threshold),
        "iterative_pruning_cpss_min_stable_features": int(iterative_pruning_bounded_cpss_min_stable_features),
        "iterative_pruning_cpss_min_jaccard": float(iterative_pruning_bounded_cpss_min_jaccard),
        "iterative_pruning_cpss_max_score_drop": float(iterative_pruning_bounded_cpss_max_score_drop),
        "iterative_pruning_cpss_stable_feature_count": 0,
        "iterative_pruning_cpss_overlap_jaccard": float("nan"),
        "iterative_pruning_cpss_overlap_recall": float("nan"),
        "iterative_pruning_cpss_base_score": float(best_score) if np.isfinite(best_score) else float("nan"),
        "iterative_pruning_cpss_overlay_score": float("nan"),
        "iterative_pruning_cpss_overlay_score_evaluated": False,
        "iterative_pruning_cpss_freq_q50": float("nan"),
        "iterative_pruning_cpss_freq_q90": float("nan"),
    }
    if runtime_bounded and iterative_pruning_bounded_use_cpss_overlay:
        cpss_meta = cpss_overlay_fn(
            X_pool=X_pool,
            y_arr=y_arr,
            target=target,
            base_selected=selected_local,
            base_relevance=base_relevance,
            base_score=best_score,
        )
        if bool(cpss_meta.get("iterative_pruning_cpss_overlay_score_evaluated", False)):
            evaluations += 1
        if bool(cpss_meta.get("iterative_pruning_cpss_switch_applied", False)):
            selected_local = np.asarray(cpss_meta.get("iterative_pruning_cpss_selected_local", selected_local), dtype=int)
            overlay_score = float(cpss_meta.get("iterative_pruning_cpss_overlay_score", float("nan")))
            if np.isfinite(overlay_score):
                best_score = float(overlay_score)

    selected_local = np.array(
        sorted(set(int(i) for i in np.asarray(selected_local, dtype=int).tolist() if 0 <= int(i) < n_pool)),
        dtype=int,
    )
    if selected_local.size > target:
        selected_local = selected_local[:target]
    if selected_local.size < target:
        fill_order = np.argsort(base_relevance)[::-1]
        present = set(int(i) for i in selected_local.tolist())
        for idx in fill_order.tolist():
            idx_i = int(idx)
            if idx_i in present:
                continue
            selected_local = np.append(selected_local, idx_i)
            present.add(idx_i)
            if selected_local.size >= target:
                break
    selected_local = np.asarray(selected_local[:target], dtype=int)
    selected_global = np.asarray(pool_idx[selected_local], dtype=int)

    corr_local = corr_pool[np.ix_(selected_local, selected_local)]
    selected_mean_corr = np.mean(corr_local, axis=1) if corr_local.size else np.zeros(selected_local.size, dtype=float)
    selected_signal = 0.65 * normalize_fn(base_relevance[selected_local]) + 0.35 * (
        1.0 - normalize_fn(selected_mean_corr)
    )
    if bool(cpss_meta.get("iterative_pruning_cpss_switch_applied", False)):
        stable_boost = np.zeros(selected_local.size, dtype=float)
        stable_boost[:] = 1.0
        selected_signal = 0.60 * normalize_fn(base_relevance[selected_local]) + 0.25 * (
            1.0 - normalize_fn(selected_mean_corr)
        ) + 0.15 * stable_boost

    all_scores = np.zeros(n_features, dtype=float)
    all_scores[pool_idx[selected_local]] = np.asarray(selected_signal, dtype=float)
    if np.max(all_scores) <= 0:
        all_scores[selected_global] = np.linspace(1.0, 0.5, num=max(1, selected_global.size))
    all_scores = normalize_fn(all_scores)

    selected_order = np.argsort(all_scores[selected_global])[::-1]
    selected_global = selected_global[np.asarray(selected_order, dtype=int)]

    results = {
        "selected_indices": np.asarray(selected_global, dtype=int),
        "scores": {int(idx): float(all_scores[int(idx)]) for idx in np.asarray(selected_global, dtype=int).tolist()},
        "method": method_name,
        "iterative_pruning_initial_size": int(init_size),
        "iterative_pruning_final_size": int(selected_global.size),
        "iterative_pruning_rounds": int(rounds),
        "iterative_pruning_pool_size": int(pool_idx.size),
        "iterative_pruning_pool_factor": float(iterative_pruning_pool_factor),
        "iterative_pruning_redundancy_weight": float(iterative_pruning_redundancy_weight),
        "iterative_pruning_min_improvement": float(iterative_pruning_min_improvement),
        "iterative_pruning_max_cumulative_loss": float(iterative_pruning_max_cumulative_loss),
        "iterative_pruning_cumulative_delta": float(cumulative_delta),
        "iterative_pruning_max_rounds": int(iterative_pruning_max_rounds),
        "iterative_pruning_removed_features": list(removed_features),
        "iterative_pruning_removed_deltas": list(removed_deltas),
        "iterative_pruning_best_score_final": float(best_score) if np.isfinite(best_score) else float("nan"),
        "iterative_pruning_evaluations": int(evaluations),
        "iterative_pruning_stop_reason": str(stop_reason),
        "iterative_pruning_runtime_bounded": bool(runtime_bounded),
        "iterative_pruning_runtime_sec": float(max(0.0, perf_counter() - start_t)),
        "iterative_pruning_candidate_budgets": list(per_round_candidate_budget),
        "iterative_pruning_corr_estimator": str(corr_estimator),
        "iterative_pruning_corr_shrinkage": float(corr_shrinkage),
        "iterative_pruning_corr_n": int(X_pool.shape[0]),
        "iterative_pruning_corr_p": int(X_pool.shape[1]),
        "iterative_pruning_classes": int(n_classes),
        "iterative_pruning_min_class_count": int(min_class_count),
        "iterative_pruning_imbalance_ratio": float(imbalance_ratio),
    }
    results.update(pareto_meta)
    results.update(cpss_meta)
    if runtime_bounded:
        results.update(
            {
                "iterative_pruning_bounded_prefilter_cap": int(iterative_pruning_bounded_prefilter_cap),
                "iterative_pruning_bounded_candidate_fraction": float(
                    iterative_pruning_bounded_candidate_fraction
                ),
                "iterative_pruning_bounded_min_candidates": int(iterative_pruning_bounded_min_candidates),
                "iterative_pruning_bounded_max_evaluations": int(iterative_pruning_bounded_max_evaluations),
                "iterative_pruning_bounded_max_runtime_seconds": float(
                    iterative_pruning_bounded_max_runtime_seconds
                ),
                "iterative_pruning_bounded_enable_class_gating": bool(
                    iterative_pruning_bounded_enable_class_gating
                ),
                "iterative_pruning_bounded_multiclass_scale": float(
                    iterative_pruning_bounded_multiclass_scale
                ),
                "iterative_pruning_bounded_imbalance_trigger": float(
                    iterative_pruning_bounded_imbalance_trigger
                ),
                "iterative_pruning_bounded_imbalance_scale": float(
                    iterative_pruning_bounded_imbalance_scale
                ),
            }
        )
    return results, {i: float(all_scores[i]) for i in range(n_features)}
